# Usage: u_ys = msnODE2 (xs, A, B, f, C, g, N, n, s, M1, M2, tau)
#
# See Papers/MSNODE/notes.tm for full documentation
# See msnODE.m for previous implementation.
## Main difference is that we also compute a grading left orthogonal transform
## for the interface equations. This *should* increase numerical reliability.
#
# Solves an ODE of the form
# A(x) u'(x) + B(x) u(x) = f(x), ys[1] < x < ys[Np], x != ys[i]
# CL u(ys[1]) = gL
# u(ys[i]+) - u(ys[i]-) = 0
# CR u(ys[Np]) = gR
# where A, B are qxq
#
# Patch MSN parameters
# ns --- number of points in local patch. size(ns) = Np-1
# s --- Sobolev parameter
# M1 -- Factor multiplying mesh norm to give polynomial order
#
# A, B, f -- are matrix-valued functions
# -- A & B must take a single point
# -- f must take an *array* of points and use them as a row index
#
# For single precision make sure CL is single precision
#
# $Id$
#

function us = msnODE2 (q, A, B, f, CL, CR, gL, gR, s, M1, ys, ns)
  ## 
  realType = class (CL);
  # Basic sizes
  nrhs = columns (gL);
  Np = max (size (ys));
  #
  Iq = eye (q, realType);
  I2q = eye (2*q, realType);
  #
  # Sparse matrix set-up
  spN = (Np-2)*q; # u(ys(2)), ..., u(ys(Np-1))
  spM = 2*spN;
  NNZ = 2*q*(q+1)/2 + (Np-3)*2*q*(2*q+1)/2;
  spTrip = zeros (NNZ, 3, realType);
  spInd = 1;
  spRHS = zeros (spM, nrhs, realType);
  mps = 0; # start row for current patch
  nps = 0; # start col for current patch
  #
  # Loop over patches and fill spTrip
  #
  ## Left-most patch
  ### Generate grid on [ys(1),ys(2)]
  m = cast (ns(1), realType);	# no. of points in grid
  n = 2*(m+2);			# degree of Chebyshev polynomial
  ord = [0:n-1];
  zs = cheb_zeros (m);
  xs = to_interval (ys(1), ys(2), zs);
  ### Form coeff matrix
  CVp = kron ((2 / (ys(2)-ys(1))) * cheb_deriv (zs, ord) , Iq);
  CV = kron (cheb (zs, ord), Iq);
  #### Vu2 = A(xs) * CVp + B(xs) * CV;
  for i=1:m
    CVp(1+(i-1)*q:i*q,:) = A(xs(i)) * CVp(1+(i-1)*q:i*q,:);
    CV(1+(i-1)*q:i*q,:) = B(xs(i)) * CV(1+(i-1)*q:i*q,:);
  endfor
  Vu = [(CL * kron (cheb([-1], ord), Iq)); (CVp + CV)];
  Ds = kron ((1+ord) .^ s, ones (1, q, realType));
  Dsi = diag (1 ./ Ds);
  Vu = Vu * Dsi;
  ### Coupled equations
  Vc = kron (cheb([1], ord), Iq) * Dsi;
  ## Coupled equations for interval on right
  m1 = cast (ns(1+1), realType);
  n1 = 2 * m1;
  ord1 = [0 : n1-1];
  Vcr = kron (cheb ([-1], ord1), Iq);
  Ds1 = kron ((1+ord1) .^ s, ones (1, q, realType));
  Ds1i = diag (1 ./ Ds1);
  Vcr = Vcr * Ds1i;
  ## Compute joint left singular vectors
  Vclr = [Vc Vcr];
  colnrms = norm (Vclr, Inf, "columns");
  [cn, piJ] = sort (colnrms, "descend");
  Vclr = Vclr(:,piJ);
  [u1, _, _] = svd (Vclr);
  ## Pre-condition from left
  Vc = u1' * Vc;
  Vcr = u1' * Vcr;
  ## RHS for uncoupled equations
  b = [gL; f(xs)];
  ##  Vc = Vc * v; u will left pre-condition b
  [u, l, Vc] = rrulv2 (Vu, b, Vc);
  m1 = rows (b);
  assert (m1 == m*q + rows(CL));
  a1 = l \ (u' * b);
  c2 = Vc(:,1:m1) * a1;
  Vc2 = Vc(:,m1+1:end);
  [_, l1t] = qr (Vc2', 0);
  l1 = (l1t') \ Iq;
  c2 = (l1t') \ c2;
  ### Copy into spTrip
  for j=1:q
    for i=j:q
      spTrip(spInd,:) = [mps+i, nps+j, l1(i,j)];
      spInd = spInd + 1;
    endfor
  endfor
  #
  ### Copy into spRHS
  spRHS(mps+1:mps+q,:) = c2;
  # Update mps & nps for next patch
  mps = mps+q;
  nps = nps+0;
  #
  ## Loop over interior patches
  for p=2:Np-2
    ### Generate grid on [ys(p),ys(p+1)]
    m = cast (ns(p), realType);
    n = 2*m;
    ord = [0:n-1];
    zs = cheb_zeros (m);
    xs = to_interval (ys(p), ys(p+1), zs);
    ### Form coeff matrix
    CVp = kron (2 * cheb_deriv (zs, ord) / (ys(p+1)-ys(p)), Iq);
    CV = kron (cheb (zs, ord), Iq);
    #### Vu = A(xs) * CVp + B(xs) * CV;
    for i=1:m
      CVp(1+(i-1)*q:i*q,:) = A(xs(i)) * CVp(1+(i-1)*q:i*q,:);
      CV(1+(i-1)*q:i*q,:) = B(xs(i)) * CV(1+(i-1)*q:i*q,:);
    endfor
    Vu = CVp + CV;
    Ds = kron([1:n] .^ s, ones (1, q, realType));
    Dsi = diag (1./Ds);
    Vu = Vu * Dsi;
    ### Coupled equations
    # Vc = [kron(cheb([-1], ord), Iq);
    #       kron(cheb([ 1], ord), Iq)] * Dsi;
    Vcl1 = Vcr;			# Old Vcr already scaled & left pre-conditioned
    Vcr1 = kron (cheb([1], ord), Iq) * Dsi;
    ## Coupled equations of next interval
    m1 = cast (ns(p+1), realType);
    n1 = 2 * m1;
    ord1 = [0 : n1-1];
    Vcr = kron (cheb ([-1], ord1), Iq);
    Ds1 = kron ((1+ord1) .^ s, ones (1, q, realType));
    Ds1i = diag (1 ./ Ds1);
    Vcr = Vcr * Ds1i;
    ## Compute joint left singular vectors
    Vclr = [Vcr1 Vcr];
    colnrms = norm (Vclr, Inf, "columns");
    [cn, piJ] = sort (colnrms, "descend");
    Vclr = Vclr(:,piJ);
    [u1, _, _] = svd (Vclr);
    ## Pre-condition from left
    Vcr1 = u1' * Vcr1;
    Vcr = u1' * Vcr;
    ## Local coupled equations
    Vc = [Vcl1; Vcr1];		# scaled and left pre-conditioned
    ## RHS for uncoupled equations
    b = f(xs);
    ## Vc = Vc * v; u will left pre-condition b
    [u, l, Vc] = rrulv2 (Vu, b, Vc);
    m1 = rows(b);
    assert (m1 == m * q);
    a1 = l \ (u' * b);
    c2 = Vc(:,1:m1) * a1;
    Vc2 = Vc(:,m1+1:end);
    [_, l1t] = qr (Vc2', 0);
    l1 = l1t' \ I2q;
    c2 = l1t' \ c2;
    ### Copy into spTrip
    for j = 1 : 2*q
      for i = j : 2*q
	spTrip(spInd,:) = [mps+i, nps+j, l1(i,j)];
	spInd = spInd + 1;
      endfor
    endfor
    #
    ### Copy into spRHS
    spRHS(mps+1:mps+2*q,:) = c2;
    #
    ### Update mps & nps for next patch
    mps = mps + 2*q;
    nps = nps + q;
    #
  endfor
  #
  ## Right-most patch
  #
  ### Generate grid on [ys(Np-1),ys(Np)]
  m = cast (ns(Np-1), realType);
  n = 2*m;
  ord = [0:n-1];
  zs = cheb_zeros (m);
  xs = to_interval (ys(Np-1), ys(Np), zs);
  ### Form coeff matrix
  CVp = kron (2*cheb_deriv(zs, ord)/(ys(Np)-ys(Np-1)), Iq);
  CV = kron (cheb(zs, ord), Iq);
  #### Vu1 = A(xs) * CVp + B(xs) * CV;
  for i=1:m
    CVp(1+(i-1)*q:i*q,:) = A(xs(i)) * CVp(1+(i-1)*q:i*q,:);
    CV(1+(i-1)*q:i*q,:) = B(xs(i)) * CV(1+(i-1)*q:i*q,:);
  endfor
  Vu = [CVp + CV; CR * kron(cheb([1], ord), Iq)];
  Ds = kron([1:n].^s, ones(1,q,realType));
  Dsi = diag(1./Ds);
  Vu = Vu * Dsi;
  ### Coupled equations
  # Vc = kron(cheb([-1], ord), Iq) * Dsi;
  Vc = Vcr;			# Old Vcr already scaled & pre-conditioned
  ## RHS for uncoupled equations
  b = [f(xs); gR];
  ## Vc = Vc * v;
  [u, l, Vc] = rrulv2 (Vu, b, Vc);
  m1 = rows(b);
  assert (m1 == rows(CR) + m*q);
  a1 = l \ (u' * b);
  c2 = Vc(:,1:m1) * a1;
  Vc2 = Vc(:,m1+1:end);
  [_, l1t] = qr(Vc2', 0);
  l1 = l1t' \ Iq;
  c2 = l1t' \ c2;
  ### Copy into spTrip
  for j=1:q
    for i=j:q
      spTrip(spInd,:) = [mps+i, nps+j, l1(i,j)];
      spInd = spInd + 1;
    endfor
  endfor
  #
  ### Copy into spRHS
  spRHS(mps+1:mps+q,:) = c2;
  #
  ### Update mps & nps for next patch
  mps = mps + q;
  nps = nps + q;
  #
  # Sanity check
  assert (mps == (Np-2)*2*q);
  assert (nps == (Np-2)*q);
  #
  assert (spInd-1, NNZ);
  # Form sparse matrix
  spS = sparse(spTrip(:,1), spTrip(:,2), spTrip(:,3));
  # spy(spS);
  #
  # Solve sparse least-squares problem
  [qTc, spR] = qr(spS, spRHS);
  us = spR \ qTc;
  #
  # Return answer
  #
endfunction

# Local Variables: #
# mode: octave     #
# End:             #
