## Usage [zzs, us] = msnPDE(Geom, igrid, q, p, nrhs, A1, A2, B, f, C, g, \
##                     ss, ns, nse, M1)
## 
## Geom.Pts ---- Pts(i,:) = [xi yi] -- all points that matter to geometry
## 
## Geom.Lines -- Lines{i} = [i,j,k,..] such that Pts(i), Pts(j), Pts(k), ...
##          appear in that order on Lines{i}
## 
## Geom.Edges -- Edges(i,:) = [li,i1,i2] == Pts(i1) <--> Pts(i2)
##          Vertices lie on Lines{li}
## 
## Geom.Quads -- Quads{i} = [e1, e2, e3, e4, e5, e6, ...],
##          ei == +- Edges{i}
##          Edges are assumed to be in anti-clockwise direction.
##          Note: A Quad is a polygon with 4 non co-linear edges. Vertices
##                have to be extracted manually. Two edges are co-linear
##                iff they belong to the same line.
## 
## Geom.BEdges --- Edges{BEdges(i)} is a boundary edge
## 
## Geom.IEdges --- Edges{IEdges(i)} is an interior edge
## 
## igrid -- igrid{i} = [z1; z2; ...]: grid points on i-th interior edge
##          length(igrid) == rows(Geom.IEdges)
## 
## The PDE is of the form
## 
## sum_{i=1}^2 A_i(x,y) partial_i u + B(x,y) u = f(x,y)
## 
## -- u ---- p x nrhs
## -- A_i(j,x) -- q x p
## -- B(j,x) ---- q x p
## -- f(j,x) ---- q x nrhs
## 
##    In all of the above j denotes the index number of the quadrilateral in
##    which x lies.
## 
## Boundary condition is of the form
## If z in E_i common to two Quads
## -- u(z+) - u(z-) = 0
## If z in E_i belonging to only one Quad (and hence boundary edge)
## -- C(z) u(z) = g(z)
##    C(j,z) -- q_i x p, where j is the edge in which z lies
##    g(j,z) -- q_i x nrhs
## 
## ss(i) ---- Sobolev parameters for Quads(i); single for single precision
## ns(i,:) -- 2D grid size for Quads{i}
## nse(i) --- 1D grid size for Edges(i); == length(igrid{i}) for interior edges
## M1 ------- M1 * mesh_norm == MSN order
## 
## zzs --- all interface grid points in the same order as us
## us ---- solution on zzs
## 
## $Id: msnPDE.m,v 1.3 2012/06/02 03:10:35 shiv Exp shiv $
## 

function [zzs, us] = msnPDE (Geom, igrid, q, p, nrhs, A1, A2, B, f, C, g, \
			     ss, ns, nse, M1)
  ## 
  realType = class(ss);
  ## 
  Np = length (Geom.Quads);	# no. of patches
  NBe = length (Geom.BEdges);	# no. of boundary edges
  NIe = length (Geom.IEdges);	# no. of interior edges
  assert (NIe, length (igrid));	# check if enough grids were provided
  ## igrid slots are 1-1 with Geom.IEdges
  ## 
  ## Count equations in each patch, and unknowns
  ## Equations/patch == no. of unknowns on boundaries
  ## Unknowns == no. of unknowns on all (internal) boundaries
  ## NNZ = sum_patches l*(l+1)/2 where l is the number of equations
  ##          in one patch
  #
  spM = 0; 			# no. of equations
  NNZ = 0;			# no. of non-zeros
  #
  for i = 1 : Np
    Edgesi = Geom.Quads{i};		# edges in i-th quadrilateral
    bi = length (Edgesi);
    lM = 0;
    for j = 1 : bi
      edgej = abs(Edgesi(j));
      if any (Geom.IEdges == edgej) # if j-th edge is an interior edge
	k = find (Geom.IEdges == edgej); # interior edge index no.
	nej = length (igrid{k});	# retrive grid
	spM = spM + p*nej;
	lM = lM + p*nej;
      endif
    endfor
    NNZ = NNZ + lM * (lM + 1)/2;
  endfor
  #
  ## Assign column numbers to each unknown. Each grid point has p
  ## unknowns.
  spN = 0;			# no. of unknowns
  # Array that holds starting column index for each interior edge
  UInd = zeros (NIe, 1);
  #
  for i = 1 : NIe
    UInd(i) = spN+1;
    nei = length (igrid{i});
    spN = spN + p*nei;
  endfor
  #
  ## Create zzs
  zzs = zeros (round(spN/p), 2);
  zzsi = 0;
  for i = 1 : NIe
    zsei = igrid{i};
    nei = length (zsei);
    zzs(zzsi+1 : zzsi+nei, :) = zsei;
    zzsi = zzsi + nei;
  endfor
  assert (zzsi*p, spN);
  #
  ## Create sparse triplet  
  spTrip = zeros (NNZ, 3, realType);
  spRHS = zeros (spM, nrhs, realType);
  spInd = 1;			# next free triplet
  mps = 0;			# no. of preceeding sparse equations
  ##
  ## Standard square  
  PQ = [-1 -1;
	 1 -1;
	 1  1;
	-1  1];
  #  
  Ip = eye (p, realType);
  #
  ## Loop over patches filling spTrip and spRHS
  ## 
  [tt, tu, ts] = cputime ();
  for patch = 1 : Np
    ## 
    ## Extract vertices of current quad
    cur_edges = Geom.Quads{patch};
    ne = length (cur_edges);	# no. of edges
    vertices = [];
    prev_line = -1;		# none
    for i = 1 : ne
      cur_line = Geom.Edges(abs(cur_edges(i)),1);
      if cur_line != prev_line	# new edge of quad
	## Get head of edge
	if cur_edges(i) > 0
	  vert1 = Geom.Edges(cur_edges(i), 2);
	else
	  vert1 = Geom.Edges(abs(cur_edges(i)), 3);
	endif
	#
	if (any (vertices == vert1))
	  patch = patch
	  edge_no = i
	  cur_line = cur_line
	  prev_line = prev_line
	  cur_edges = cur_edges
	  vertices = vertices
	  vert1 = vert1
	  error ("Huh?");		# how did it repeat?
	else
	  vertices = [vertices vert1];	# add vertex
	endif
      endif
      prev_line = cur_line;
    endfor
    assert (length (vertices), 4);
    #
    ## Compute map from quad to [-1,1]x[-1,1] and back
    XY = Geom.Pts(vertices, :);
    Af = quad2quad (XY, PQ);
    # Ai = Af \ eye (3, realType);
    #
    ## Generate interior grid for patch; zs(n2*(i-1)+j,:) = [zs1(i), zs2(j)]
    n1 = ns(patch, 1);
    n2 = ns(patch, 2);
    zs = [(kron (cheb_zeros (n1), ones (n2, 1))), \
	  (kron (ones (n1, 1), cheb_zeros (n2)))];
    xs = toQuad (XY, zs);
    assert (rows(xs), n1*n2);
    ## 
    ## Find boundary edges and no. of equations on those edges
    bedges = [];
    iedges = [];
    mbd = 0;		# tot. no. of boundary conditions for this quad
    mint = 0;		# tot. no. of unknowns on int. edges for this quad
    for i = 1 : ne
      ei = abs(cur_edges(i));
      if any (Geom.BEdges == ei)	# boundary edge
	vt1 = Geom.Pts(Geom.Edges(ei, 2),:);
	vt2 = Geom.Pts(Geom.Edges(ei, 3),:);
	midpt = (vt1 + vt2) / 2;
	mbd = mbd + nse(ei) * rows (C (ei, midpt));
	bedges = [bedges ei];
      else			# interior edge
	iedges = [iedges ei];
	mint = mint + nse(ei) * p;
      endif
    endfor
    bem = length (bedges);
    iem = length (iedges);
    assert (bem+iem, ne);     # Check if we retained all edges.
    #
    ## Prepare to generate uncoupled equations
    assert (columns(xs), 2);
    m = rows (xs);		# no. of grid points
    N1 = round (M1*n1);
    N2 = round (M1*n2);
    Vu = zeros (mbd+m*q, N1*N2*p, realType);
    RHS = zeros (mbd+m*q, nrhs);
    #
    ## Degrees of 2D Chebyshev polynomials
    ord = [(kron ([0 : N1-1]', ones(N2, 1))), \
	   (kron (ones(N1, 1), [0 : N2-1]'))];
    assert (columns (ord), 2);
    #
    cur_row = 1;
    #
    ## Generate boundary uncoupled equations --- mbd of them
    for i = 1 : bem
      ei = bedges(i);
      # li = Edges(ei, 1); not needed
      v1i = Geom.Edges(ei, 2);
      v2i = Geom.Edges(ei, 3);
      nei = nse(ei);
      zsei = cheb_zeros (nei);
      xsei = toEdge (Geom.Pts(v1i, :), Geom.Pts(v2i, :), zsei);
      zsei = fromQuad (XY, xsei);
      zsei = roundToEdge (zsei);
      for j = 1 : nei
	# Fill-in V
	Cj = C (ei, xsei(j,:));
	mj = rows (Cj);
	Vu(cur_row:cur_row+mj-1, :) =  Cj * kron (cheb2D (zsei(j,:), ord), Ip);
	# Fill-in RHS
	RHS(cur_row:cur_row+mj-1, :) = g (ei, xsei(j,:));
	#
	cur_row = cur_row + mj;
      endfor
      #
    endfor
    assert (cur_row, mbd+1);	# sanity check
    ##
    ## Generate interior uncoupled equations --- q*n1*n2 of them
    for i = 1 : n1*n2
      # Cheb-Vandermonde matrices
      Vi = cheb2D (zs(i,:), ord);
      [V1i, V2i] = cheb2D_deriv (zs(i,:), ord);
      V1i = kron (V1i, Ip);
      V2i = kron (V2i, Ip);
      phip = persp_deriv (Af, xs(i,:)); # [phi1_x, phi1_y; phi2_x, phi2_y]
      # PDE coeffs
      A1i = A1 (patch, xs(i,:));
      A2i = A2 (patch, xs(i,:));
      Bi = B (patch, xs(i,:));
      fi = f (patch, xs(i,:));
      # Fill-in
      Vu(cur_row : cur_row + q - 1, :) = \
	  A1i * (phip(1,1) * V1i + phip(2,1) * V2i) + \
	  A2i * (phip(1,2) * V1i + phip(2,2) * V2i) + \
	  Bi * kron (Vi, Ip);
      RHS(cur_row : cur_row + q - 1, :) = fi;
      # Update row index
      cur_row = cur_row + q;
      #
    endfor
    assert (cur_row = mbd + q * n1 * n2 + 1); # sanity check
    ## ###### DEBUG ######
    # dbg_fs = sample2D (p, (@(xy) [xy(1), 0]), xs);
    # dbg_ac = msn_interp2D (XY, xs, dbg_fs, ss(patch), [N1 N2]);
    # dbg_ac = reshape(dbg_ac', N1*N2*p, 1);
    # dbg_err = norm (Vu * dbg_ac - RHS, Inf) / \
    # 	(norm (Vu, Inf) * norm (dbg_ac, Inf) + norm (RHS, Inf))
    ## ####### END #######
    ## 
    ## Scale with Sobolev weights
    Ds = (1 + ord(:,1)+ord(:,2)) .^ ss(patch);
    Ds = kron (Ds, ones (p, 1));
    Dsi = diag (1 ./ Ds);
    Vu = Vu * Dsi;
    ## 
    ## Generate coupled equations
    Vc = zeros (mint, N1*N2*p, realType);
    cur_row = 1;
    for i = 1 : iem
      ei = iedges(i);		# i-th interior edge global index
      k = find (Geom.IEdges == ei);
      xsei = igrid{k};		# grid on i-th interior edge
      mi = rows (xsei);
      assert (columns (xsei), 2);
      zsei = fromQuad (XY, xsei); # cannot use fromEdge!
      zsei = roundToEdge (zsei);  # make sure it did not fall outside!
      Vc(cur_row : cur_row + p*mi - 1, :) = kron (cheb2D (zsei, ord), Ip);
      cur_row = cur_row + p*mi;
    endfor
    #
    ## ###### DEBUG #######
    # dbg_m = length (igrid{1});
    # dbg_ue = sample2D (2, @(xy) [xy(1) , 0], igrid{1});
    # dbg_ue = reshape(dbg_ue', dbg_m * p, 1);
    # dbg_err_edg = norm (Vc * dbg_ac - dbg_ue, Inf) / \
    # 	(norm (Vc, Inf) * norm (dbg_ac, Inf) + norm (dbg_ue, Inf))
    ## ####### END ########
    Vc = Vc * Dsi;
    ## 
    ## Compress uncoupled equations
    # [u, l, Vc] = rrulv (Vu, Vc); # Vc <- Vc * v
    [u, l, Vc] = rrulv2 (Vu, RHS, Vc); # Vc <- Vc * v
    a1 = l \ (u' * RHS);
    ## ####### DEBUG ########    
    # dbg_ac = diag(Ds) * dbg_ac;
    # [dbg_u, dbg_l, dbg_act] = rrulv (Vu, dbg_ac');
    # dbg_ac = dbg_act';
    # dbg_m = rows (a1);
    # [a1, dbg_ac(1:dbg_m,:)]
    # dbg_err_rrulv = norm(a1 - dbg_ac(1:dbg_m,:), Inf) / norm(dbg_ac, Inf)
    ## ######## END #########
    ## 
    ## Compress coupled equations
    m1 = rows (RHS);
    c2 = Vc(:,1:m1) * a1;
    Vc2 = Vc(:,m1+1:end);
    [_, l1t] = qr (Vc2', 0);
    l1 = l1t' \ eye (mint, realType);
    c2 = l1t' \ c2;
    ## 
    ## Fill-in spTrip.
    cur_j = 0;			# local column index
    for iej = 1 : iem
      ei = iedges(iej);
      k = find (Geom.IEdges == ei);
      nps = UInd(k) - 1;
      nej = length (igrid{k});
      for j = 1 : p*nej
	for i = cur_j+j : mint
	  spTrip(spInd,:) = [mps+i, nps+j, l1(i,cur_j+j)];
	  spInd = spInd + 1;	  
	endfor
      endfor
      cur_j = cur_j + p*nej;
    endfor
    ## 
    ## Fill in spRHS.
    spRHS(mps+1:mps+mint,:) = c2;
    #
    ## Update mps for next patch
    mps = mps + mint;
    #
  endfor			# for patch = 1:Np
  [tt1, tu1, ts1] = cputime ();
  cpu_per_patch = (tt1-tt) / Np
  #
  assert (spInd-1, NNZ);	# sanity check
  ## For sparse matrix
  spS = sparse (spTrip(:,1), spTrip(:,2), spTrip(:,3));
  #
  ## Solve sparse LS problem
  [tt, tu, ts] = cputime ();
  [qTc, spR] = qr (spS, spRHS);
  us = spR \ qTc;
  [tt1, tu1, ts1] = cputime ();
  cpu_sparse_solve = tt1-tt
  #
endfunction

# Local Variables: #
# mode: octave     #
# End:             #
