% -*- mode: octave -*-
%
% Test of msnODE.tm
%
% $Id$

%%%%%%%%%%%%%%%%%%%%%%%%

% ODE parameters
a = 0.0 % left-end of interval
b = 1.0 % right-end of interval
xs = [a b] % The boundary points
q = 2 % order of system
q1 = 1 % number of left boundary conditions
q2 = q-q1 % number of right boundary conditions

% ODE is of the form: A(x)*u'(x) + B(x)*u(x) = f(x), a < x < b.
% C(a) * u(a) = g(a)
% C(b) * u(b) = g(b)

function B = coeff0 (x)
  B = [cosh(x), sinh(x); sinh(x), cosh(x)];
endfunction

function A = coeff1 (x)
  A = [cos(x), -sin(x); sin(x), cos(x)];
endfunction

% [soln; soln']
function u = soln (x)
  u = [sin(4*pi*x) * x^2 * (1-x)^2,          \
       -2 * (1-x) * x^2 * sin(4*pi*x) +      \
       2 * (1-x)^2 * x * sin(4*pi*x) +       \
       4 * pi * (1-x)^2 * x^2 * cos(4*pi*x); \
       cos(5*pi*x) * x^3 * (1-x)^3,            \
       -5 * pi * (1-x)^3 * x^3 * sin(5*pi*x) - \
       3 * (1-x)^2 * x^3 * cos(5*pi*x) +       \
       3 * (1-x)^3 * x^2 * cos(5*pi*x)];
endfunction

function f = rhs (x)
  u = soln(x);
  f = coeff0(x) * u(:,1) + coeff1(x) * u(:,2);
endfunction

function C = bc_coeff (x)
  if (abs(x) < sqrt(eps))
    C = [1 0];
  elseif (abs(1-x) < sqrt(eps))
    C = [0 1];
  else
    assert (1 == 0);
  endif
endfunction

function g = bc_rhs (x)
  C = bc_coeff(x);
  if (abs(x) < sqrt(eps))
    u = soln(0.0);
    g = C * u(:,1);
  elseif (abs(1-x) < sqrt(eps))
    u = soln(1.0);
    g = C * u(:,1);
  else
    assert (1 == 0);
  endif
endfunction

%%%%%%%%%%%%%%

% Patch MSN parameters
N = 2 % number of patches
n = 4 % points in local patch
s = 2.5 % Sobolev parameter
M1 = 2.5 % Factor multiplying mesh norm to give polynomial order
m2 = 1 % Number of initial zero terms
tau = sqrt(eps) % Weights for near zero terms

% The patches themsleves
ys = a : (b-a)/(N+1) : b;
# ys = cos((2*[N:-1:1]-1) * (pi/(2*N))); % Chebyshev grid
# ys = to_interval (a, b, ys);
# ys = cos([N-1:-1:0] * pi / (N-1)); % 2nd kind Chebyshev grid
# ys = to_interval (a, b, ys);

% Final sparse matrix parameters
sp_M = (N-1)*q % Number of rows in sparse matrix
sp_N = 2*q + (N-2)*2*q % Number of columns in sparse matrix
%sp_NNZ = 2*(q*(q+1)/2) + (N-2)*(2*q*(2*q+1)/2) % Number of non-zeros
sp_NNZ = q*(q+1) + (N-2)*q*(2*q+1)
sp_I = zeros(sp_NNZ, 1);
sp_J = zeros(sp_NNZ, 1);
sp_V = zeros(sp_NNZ, 1);

% RHS for sparse system
sp_rhs = zeros(sp_M, 1);

cur_sp_ind = 1; % current index for sparse (I,J,V) triplet

% For each patch we create the MSN matrix and compress it.
for patch=1:N

  l = ys(patch); % local patch is [l,r]
  r = ys(patch+1);
  recip_len = 2/(r-l); % scale factor for cheb_deriv

  % Set up local grid
  % xs = cos([N-1:-1:0] * pi / (N-1)); % 2nd kind Chebyshev grid
  xs = -1: 2/(n-1): 1; % equi-spaced grid on [-1,1]
  [hmin, hmax] = mesh_norm_1d (xs);
  m = ceil (M1 * pi / hmin);
  xt = to_interval (l, r, xs); % grid on [l,r]
  ident = eye(q,q);

  % Set up left end point.
  VL = zeros(0, q*m);
  fL = zeros(0, 1);
  if (patch == 1) % then left BC
    [BL, gl] = bc_left;
    VL = zeros(q1, q*m);
    for j=1:m
      c1j = cheb(j-1, -1);
      VL(:, q*j-q+1:q*j) = BL * c1j;
    endfor
    fL = gl;
  else % (1 < patch) fill-in VL and fL
    VL = zeros(q, q*m);
    % Just matching Dirichlet conditions; no ODE
    fL = zeros(q, 1);
    for j=1:m
      cij = cheb(j-1,-1);
      VL(:,q*j-q+1:q*j) = ident * cij;
    endfor
  endif

  % Set up right end point.
  VR = zeros(0, q*m);
  fR = zeros(0, 1);
  if (patch == N) % then right BC
    [BR, gr] = bc_right;
    VR = zeros(q2, q*m);
    for j=1:m
      cnj = cheb(j-1,1);
      VR(:,q*j-q+1:q*j) = BR * cnj;
    endfor
    fR = gr;
  else % (patch < N) fill-in VR and fR
    VR = zeros(q, q*m);
    % Just matching Dirichlet conditions; no ODE
    fR = zeros(q, 1);
    for j=1:m
      cij = cheb(j-1,1);
      VR(:,q*j-q+1:q*j) = ident * cij;
    endfor
  endif

  % Set up interior points
  VI = zeros(q*(n-2), q*m);
  fI = zeros(q*(n-2), 1);
  for i=2:n-1
    Ai = coeff0(xs(i));
    Bi = coeff1(xs(i));
    fI(q*(i-1)-q+1:q*(i-1),:) = rhs(xs(i));
    for j=1:m
      cij = cheb(j-1, xt(i));
      cpij = cheb_deriv(j-1, xt(i)) * recip_len;
      VI(q*(i-1)-q+1:q*(i-1), q*j-q+1:q*j) = Ai * cij + Bi * cpij;
    endfor
  endfor

  % Merge equations
  % Interface equations are always at the bottom as they are the only
  % ones that couple to other equations.
  mi = 0; % number of local equations
  mi1 = 0; % number of local equations that can be solved
  miL = 0; % number of local left interface conditions
  miR = 0; % number of local right interface conditions
  f = [];
  if (patch == 1)
    V = [VL; VI; VR];
    f = [fL; fI];
    mi = rows(V);
    mi1 = rows(f);
    miR = rows(VR);
  elseif (patch == N)
    V = [VI; VR; VL];
    f = [fI; fR];
    mi = rows(V);
    mi1 = rows(f);
    miL = rows(VL);
  else % interior patch
    V = [VI; VL; VR];
    f = fI;
    mi = rows(V);
    mi1 = rows(f);
    miL = rows(VL);
    miR = rows(VR);
  endif

  % Sobolev scaling
  D2 = diag(kron([1:m],ones(1,q)) .^ s, 0);
  D2(1:q*m2) = tau;
  D2i = diag(kron([1:m],ones(1,q)) .^ (-s), 0);
  D2i(1:q*m2) = 1/tau;

  % Compress
  [u, L, _] = rrulv (V * D2i);

  % Split L
  if (patch == 1)
    L11 = L(1:mi1, 1:mi1);
    LR1 = L(mi1+1:end, 1:mi1);
    LR2 = L(mi1+1:end, mi1+1:mi);
  elseif(patch == N)
    L11 = L(1:mi1, 1:mi1);
    LL1 = L(mi1+1:end, 1:mi1);
    LL2 = L(mi1+1:end, mi1+1:mi);
  else
    L11 = L(1:mi1, 1:mi1);
    LL1 = L(mi1+1:mi1+miL, 1:mi1);
    LL2 = L(mi1+1:mi1+miL, mi1+1:mi1+miL);
    LR1 = L(mi1+miL+1:end, 1:mi1);
    LR2 = L(mi1+miL+1:end, mi1+1:mi);
  endif

  % Local solve
  al = L11 \ (u' * f);

  % Left interface RHS
  if (1 < patch)
    fL = LL1 * al;
  else
    fL = zeros(0, 1);
  endif
  if (patch < N)
    fR = LR1 * al;
  else
    fR = zeros(0, 1);
  endif

  % Stick rest into sparse matrix and RHS
  %%% Equations nos. by patch
  %%%%% q*(patch-2)+1:++q for left interface
  %%%%% q*(patch-1)+1:++q for right interface
  %%% Variable nos. by patch
  %%%%% 1:q for patch == 1
  %%%%% q+(patch-2)*2*q+1:++2*q for 1 < patch < N
  %%%%% q+(N-2)*2*q+1:++q for patch == N
  if (patch == 1)
    % only right interface
    for i=1:q
      for j=1:i
	sp_I(cur_sp_ind) = i;
	sp_J(cur_sp_ind) = j;
	sp_V(cur_sp_ind) = LR2(i,j);
	cur_sp_ind = cur_sp_ind+1;
      endfor
      sp_rhs(i) = sp_rhs(i) - fR(i);
    endfor
  elseif (patch == N)
    % only left interface
    for i=1:q
      eqn_no = q*(N-2)+i; % got to back up to the previous right interface
      for j=1:i
	var_no = q+(N-2)*2*q+j;
	sp_I(cur_sp_ind) = eq_no;
	sp_J(cur_sp_ind) = var_no;
	sp_V(cur_sp_ind) = LL2(i,j);
	cur_sp_ind = cur_sp_ind+1;
      endfor
      sp_rhs(eqn_no) = sp_rhs(eqn_no) - fL(i);
    endfor
  else % interior patch
    % left interface
    for i=1:q
      eqn_no = q*(patch-2)+i; % back up to previous right interface
      for j=1:i
	var_no = q+(N-2)*2*q+j;
	sp_I(cur_sp_ind) = eq_no;
	sp_J(cur_sp_ind) = var_no;
	sp_V(cur_sp_ind) = LL2(i,j);
	cur_sp_ind = cur_sp_ind+1;
      endfor
      sp_rhs(eqn_no) = sp_rhs(eqn_no) - fL(i);
    endfor
    % right interface; lower-right trapezoid LR2
    for i=1:q
      eqn_no = q*(N-1)+i; % new set of equations for right interface
      for j=1:mi1+i
	var_no = q+(N-2)*2*q+j;
	sp_I(cur_sp_ind) = eq_no;
	sp_J(cur_sp_ind) = var_no;
	sp_V(cur_sp_ind) = LR2(i,j);
	cur_sp_ind = cur_sp_ind+1;
      endfor
      sp_rhs(eqn_no) = sp_rhs(eqn_no) - fR(i);
    endfor
  endif

endfor
